///////////////////////////////////////////////////////////////////////////////
// Perf-hash-map.cpp
//
// Author:  Dennis Lang 2010
// Website: http://home.comcast.net/~lang.dennis/index.html
// Date: April 2010
// Updated:  Sept 2013, Dec 2014
//
// Measure performance of various stl containers.
// Measure speed to insert, find and access slots in the following containers:
//      HashMap
//      HashMap with custom allocator
//      Map
//      Map with custom allocator
//      Hash MultiMap
//      MultiMap
//      Vector
//      Vector sorted
//      Deque
//
// The custom allocator helps nodes based containers by reusing previously freed nodes.
//
// Test flow:
//      1. Load a word list by reading a file (I typically use this source file) and break into words
//      2. Measure time to process words using each test 'container'
//      3. Processing occurs in three stages:
//          a.  Insert into container until 'groupSize' elements stored in container.
//              During insertion cycle, check for duplicates and count them.
//          b.  After insertion stage, find duplicates and count them.
//          c.  After counting 'sLookupScale * groupSize' words in stage (b)
//              clear list and go back to stage (a).
//
//  The test has been designed to improve the performance of those containers with the custom 
//  allocator.  Stage (c) clears the list and then resumes filling.  The custom allocator
//  will avoid repeated memory allocations and will reuse previously freed nodes. 
//  
//  Background threads are also executed to run the same test to stress the 'malloc' subsystem.
//  This done partially to emulate a real thread program which does work in parallel by mostly
//  to allow the container with the custom allocator to out perform the normal allocator.
//
//  Even with all of these tweaks the custom allocator does not make much of an improvement.
//  

#if 0
// Disable bounds checking.
// Becareful, all uses of STL must be compiled with same setting.
#define _SECURE_SCL 0
#endif

#define _CRT_SECURE_NO_WARNINGS

// Define USE_COUNTALLOC to count memory allocations
#define USE_COUNTALLOC
#ifdef USE_COUNTALLOC
bool showMemoryCounts = true;
#else
bool showMemoryCounts = false;
#endif

// Define TEST_BOOST if boost is available for testing. 
#define TEST_BOOST

// Define TEST_BOOST_ALLOCATOR if functional pool allocator.
// With Vs2013 and Boost 1_57 allocator crashes for x86
// and fails to link for x64.
#undef TEST_BOOST_ALLOCATOR

// Visual Studio preprocessors
#if defined(_M_AMD64) || defined(_M_X64)
  const char* BuildTarget = "x64";
#else
  const char* BuildTarget = "x86";
#endif

#ifdef _UNICODE
  const char* CharSet = "Unicode";
#else
  const char* CharSet = "MultiByte";
#endif

// Visual Studio preprocessors
#if defined(_MSC_VER) && !defined(COMPILERVERSION)
 #if _MSC_VER==1500
  const char* CompilerVersion = "VS2008";
 #elif _MSC_VER==1600
  const char*  CompilerVersion = "VS2010";
 #elif _MSC_VER==1700
  const char*  CompilerVersion = "VS2012";
 #elif _MSC_VER==1800
  const char*  CompilerVersion = "VS2013";
 #else
  const char*  CompilerVersion = "VS2005";
 #endif
#else
 #ifdef COMPILERVERSION
  const char*  CompilerVersion = str(COMPILERVERSION);
 #else
  const char*  CompilerVersion = "unknown";
 #endif
#endif

// Compile with /DHARDWARE="IntelPentium@1.2Ghz" to set hardware type
// or set in Property C+= prerprocessor definitions.
#ifdef HARDWARE
  const char* HardwareDesc = str(HARDWARE);
#else
 // const char* HardwareDesc = "Intel(R) T2400@ 1.83Ghz";
 const char* HardwareDesc = "Intel Xeon E5440@ 2.83Ghz";
#endif


#ifdef BUILDDESC                            
 const char* BuildDesc = str(BUILDDESC);
#else
 #ifdef _DEBUG
  const char* BuildDesc = "Debug";
 #else
  const char* BuildDesc = "Release";
 #endif
#endif
// ----- Default constants ------

// After scanning the container groupSize * sLookupScale, empty container and refill
// The refill helps leverage the custom allocator.
const int sLookupScale = 4;  

// Default number of background threads to run to stress the malloc subsystem.
// This will emulate a real program and increase the performance of the custom allocator
// which does fewer allocations.
const int sThreadCnt = 4;
const int sMaxThreads = 10;

// Default number of times to repeat test cycle to improve accuracy of results.
const int sRepeatCnt = 4;
  
// ----- Sample Results ------
 
// Jan 2011
// vs2010/x64 dual quad core Intel Xeon 3.20GHz
//
// Input Word Analysis:
//      40348 words
//        967 unique words
//         41 average #duplicates
//          6 average length
//         39 maximum length
//
// Threads:4, Repeat:4
// Results reported as ratio of elapsedTicks/fastestTicks(11542280.00):
// Group Size,      Hash, HashAlloc,      Map,   MapAlloc, HashMulti,  MultiMap,    Vector,SortVector,     Deque
//        100,     1.27 ,     1.08 ,     2.12 ,     1.86 ,     1.32 ,     2.02 ,     7.46 ,     3.39 ,    15.02
//        500,     1.42 ,     1.16 ,     2.71 ,     2.55 ,     1.38 ,     2.67 ,    25.03 ,    10.43 ,    51.41
//       1000,     1.02 ,     1.00 ,     2.55 ,     2.47 ,     1.03 ,     2.49 ,    37.13 ,    37.31 ,    74.23
// 
// Number of Memory Allocations made, malloc() call count:
// Group Size,      Hash, HashAlloc,       Map,  MapAlloc, HashMulti,  MultiMap,    Vector,SortVector,     Deque
//        100,     7004 ,       10 ,     7001 ,        7 ,     7004 ,     7001 ,       13 ,       13 ,     7351
//        500,     6460 ,       35 ,     6457 ,       32 ,     6460 ,     6457 ,       17 ,       17 ,     6548
//       1000,      972 ,       65 ,      968 ,       61 ,      972 ,      968 ,       18 ,       18 ,      976
// 
// ===============================================================================================================
// Jan 2011
//  vs2008/x64 dual quad core Intel Xeon 3.20GHz
//
// Input Word Analysis:
//      42896 words
//       1003 unique words
//         42 average #duplicates
//          6 average length
//         39 maximum length
//
// Threads:4, Repeat:4
// Results reported as ratio of elapsedTicks/fastestTicks(50498355.75):
// Group Size,      Hash, HashAlloc,      Map,   MapAlloc, HashMulti,  MultiMap,    Vector,SortVector,     Deque
//        100,     1.25 ,     1.26 ,     1.06 ,     1.00 ,     1.28 ,     1.06 ,     6.38 ,     2.40 ,    28.01
//        500,     1.41 ,     1.39 ,     1.30 ,     1.26 ,     1.45 ,     1.28 ,    25.17 ,     8.52 ,   107.02
//       1000,     1.35 ,     1.30 ,     1.36 ,     1.32 ,     1.35 ,     1.37 ,    31.22 ,    15.87 ,   130.27
// 
// Number of Memory Allocations made, malloc() call count:
// Group Size,      Hash, HashAlloc,       Map,  MapAlloc, HashMulti,  MultiMap,    Vector,SortVector,     Deque
//        100,     7105 ,       11 ,     7102 ,        8 ,     7105 ,     7102 ,       14 ,       14 ,     7598
//        500,     6925 ,       37 ,     6921 ,       33 ,     6925 ,     6921 ,       18 ,       18 ,     7074
//       1000,     6320 ,       68 ,     6316 ,       64 ,     6320 ,     6316 ,       19 ,       19 ,     6403
// 
// ===============================================================================================================
// Dec 2014
// HP xw8600 dual quad core Intel Xeon E5440 @ 2.83 GHz
// 
// Input Word Analysis :
// 40743 words
// 1409 unique words
// 28 average #duplicates
// 6 average length
// 39 maximum length
// 
// VS2013 Compiler
// x64 BuildTarget
// Release BuildType
// 
// Threads : 4, Repeat : 4
// Results reported as ratio of elapsedTicks / fastestTicks(15455703.25) :
// === [STD / STL] ===
// Group Size, HashMap, HasMpAlloc, HashMmap,   Map, MapAlloc, MultiMap, Vector, SortedVec, DequeFront, DequeBack
//        100,    1.45,       1.00,     1.37,  2.01,     1.71,     1.92,   6.08,      2.79,      12.72, 10.79
//        500,    1.75,       1.28,     1.64,  2.63,     2.43,     2.84,  27.43,      8.80,      63.33, 54.85
//       1000,    2.14,       1.85,     2.04,  3.80,     3.94,     4.22,  43.30,     18.68,     147.18, 95.10
// 
// Number of Memory Allocations made, malloc() call count :
// Group Size, HashMap, HasMpAlloc, HashMmap,  Map, MapAlloc, MultiMap, Vector, SortedVec, DequeFront, DequeBack
//        100,    7204,          8,     7204, 7201,        6,     7201,      1,          1,      7561, 7561
//        500,    6504,          8,     6504, 6501,        6,     6501,      1,          1,      6592, 6592
//       1000,    6005,          9,     6005, 6001,        6,     6001,      1,          1,      6049, 6049
// 
// 
// Threads : 4, Repeat : 4
// Results reported as ratio of elapsedTicks / fastestTicks(32960161.75) :
// === [Boost 1_57] ===
// Group Size, stdVecmap,  map, flat_map, MuliMap, Vector, DequeFront, DequeBack
//        100,      2.91, 1.01,     1.00,    1.00,   3.04,       5.61,      4.76
//        500,     16.09, 1.67,     2.13,    1.70,  15.94,      26.53,     24.38
//       1000,     23.52, 2.33,     3.21,    2.23,  25.34,      54.39,     36.53
// 
// Number of Memory Allocations made, malloc() call count :
// Group Size, stdVecmap,  map, flat_map, MuliMap, Vector, DequeFront, DequeBack
//        100,         1, 7200,        8,    7200,      1,        601,      577
//        500,         1, 6500,       10,    6500,      1,        543,      534
//       1000,         1, 6000,       11,    6000,      1,        501,      499
//
      
/////////////////////////////////////////////////////////////////////////////////

#pragma once

#include <stdio.h>
#include <assert.h>

#include <iostream>
#include <iomanip>
#include <string>
#include <algorithm>

// Portable tick counter, Cpu type and speed
#include "cycle.h"

#include <vector>
#include <map>
#include <deque>
#include <queue>
#include <stack>

// Boost
#ifdef TEST_BOOST
#define BOOST_SYSTEM_NO_DEPRECATED

#include <boost/version.hpp>
#include <boost/container/vector.hpp>
#include <boost/container/flat_map.hpp>
#include <boost/container/map.hpp>
#include <boost/container/deque.hpp>
// #include <boost/container/list.hpp>
// #include <boost/container/slist.hpp>
// #include <boost/container/set.hpp>

#ifdef TEST_BOOST_ALLOCATOR
// Use to add allocator to containers - requires libboost_thread-xxx.lib
#include <boost/pool/pool_alloc.hpp>
#include <boost/system/windows_error.hpp>
// #pragma comment(lib, "libboost_system-vc120-mt-1_57.lib")
#endif
#endif

#include <string>
#include <iostream>
#include <fstream>

typedef std::vector<std::string> Words;

#ifdef WIN32

#include <hash_map>
#include <functional>
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
typedef double Elapsed_t;
typedef HANDLE Thread_t;
typedef DWORD  Thread_return_t;
#define pthread_exit return
#ifdef min
#undef min
#undef max
#endif

#else

#include <ext/hash_map>
#include <ext/algorithm>
#include <pthread.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/resource.h>
typedef double Elapsed_t;
typedef pthread_t Thread_t;
typedef void*    Thread_return_t;
#define WINAPI 

#endif


typedef std::pair<std::string, int> PairSI;
typedef std::pair<const std::string, int> PairSIconst; 

///////////////////////////////////////////////////////////////////////////////
// Interface for Allocator
class  IAlloc
{
public:
    virtual size_t GetAllocCnt() const = 0;

    typedef size_t            size_type;

    // Optionally 
    // One shot pre-allocation value use to allocate next set of data nodes.
    // Afterwards it goes back to sAllocRate which is 16.
    virtual void SetPrealloc(size_type numElements)
    { }
};

///////////////////////////////////////////////////////////////////////////////
// Custom allocator - groups allocation in chunks of 16.
// Recyles deleted nodes by placing them on free list and using them for allocations.
// Works with STD  but not BOOST.
template <class TT>
class PoolAlloc : public IAlloc
{
    // Rate to allocate extra objects.
    static const unsigned sAllocRate = 16;

public:
    typedef size_t          size_type;
    typedef ptrdiff_t       difference_type;
    typedef TT*             pointer;
    typedef const TT*       const_pointer;
    typedef TT&             reference;
    typedef const TT&       const_reference;
    typedef TT              value_type;   

    template <class UU>    
    struct rebind
    {
        typedef PoolAlloc<UU> other;
    };
 
    // Wrapper to add singly link list pointers to track next 
    // free object and allocated pointers.
    class Wrapper
    {
    public:
        Wrapper() : pNext(0), pAlloc(0)
        { }

        Wrapper*    pNext;
        Wrapper*    pAlloc;

        // Place client data at end since its size can vary becaue
        // std::map has several internal lists of varying structure.
#if 1
        char        value[sizeof(TT)];  // reserve size for TT, avoid its constructor
#else
        TT          value;   // Safer, but slower
#endif
    };


    // Allocator shared state information.
    class SharedInfo
    {
    public:
        SharedInfo() : allocCnt(0), numToAlloc(sAllocRate), refCnt(1), pAlloc(0)
        { }
            
		~SharedInfo()
		{ FreeList(); }

        size_t   allocCnt;          // DEBUG - track number of allocations
        size_t   numToAlloc;        // # of allocations in next malloc, used to preallocate.
        int      refCnt;            // Reference counter, std::map has multiple internal list which share this info.
        Wrapper* pAlloc;            // List of all allocations

        void FreeList()
        {
            // Free any allocations (follow pAlloc link list)
            while (pAlloc != NULL)
            {
                Wrapper* pNext = pAlloc->pAlloc;
                free(pAlloc);
                pAlloc = pNext;
            }
        }
    };

    SharedInfo* pSharedInfo;
    Wrapper*    pFree;
    
    PoolAlloc() throw() : pSharedInfo(new SharedInfo()), pFree(0)
    { }

    PoolAlloc(const PoolAlloc& other) : pSharedInfo(other.pSharedInfo), pFree(0)
    { pSharedInfo->refCnt++; }

    template <class U> 
    PoolAlloc(const PoolAlloc<U>& other) : pSharedInfo((SharedInfo*)other.pSharedInfo), pFree(0)
    { pSharedInfo->refCnt++;}
    
    bool operator==(const PoolAlloc& other) 
    { return pSharedInfo == other.pSharedInfo; } 

    ~PoolAlloc () throw()
    { if (--pSharedInfo->refCnt == 0) delete pSharedInfo;  }

    // Return number of calls to malloc executed so far.
    size_t GetAllocCnt() const
    {  return pSharedInfo->allocCnt; }

    // One shot pre-allocation value use to allocate next set of data nodes.
    // Afterwards it goes back to sAllocRate which is 16.
    void SetPrealloc(size_type numToAlloc)
    {
        pSharedInfo->numToAlloc = numToAlloc;
    }

    // Pre-allocate nodes on free-list.
    void Preallocate(size_type numElements)
    {
        pSharedInfo->allocCnt++;
	
        // malloc faster than new[], avoids constructors
        Wrapper* pWrapper = (Wrapper*)malloc(numElements * sizeof(Wrapper));
        pWrapper->pNext = 0;

        // Keep track of allocated memory, wire up link list.
        pWrapper->pAlloc = pSharedInfo->pAlloc;
        pSharedInfo->pAlloc = pWrapper;

        // Split up allocation and store on free list.
        for (unsigned idx = 0; idx != numElements; idx++)
        {
            pWrapper->pNext = pFree;
            pFree = pWrapper;
            pWrapper++;
        }
    }

    pointer address(reference x) const
    { return &x; }

    const_pointer address(const_reference x) const
    { return &x; }

    pointer allocate(size_type numElements, typename PoolAlloc<TT>::const_pointer hint = 0)
    {
        if (numElements != 1)
        {
            // Pass multile element allocation directly to malloc, no special handling.
            pSharedInfo->allocCnt++;
            return static_cast<pointer>(malloc(numElements * sizeof(TT)));
        }

        if (pFree == NULL)
        {
            if (pSharedInfo->allocCnt > 4)
            {
                // Once we have complete basic setup, allocate in larger chunks.
                Preallocate(pSharedInfo->numToAlloc);
                pSharedInfo->numToAlloc = sAllocRate;
            }
            else
            {
                // During setup, allocate in small chunks.
                Preallocate(1);
            }
        }

        if (pFree == NULL)
            return 0;   // or should we throw.

        Wrapper* pWrapper = pFree; 
        pFree = pWrapper->pNext;

        pWrapper->pNext = (Wrapper*)0;

        return (TT*)&pWrapper->value;
    }

    void deallocate(pointer ptr, size_type numElements)
    {
        if (numElements != 1)
        {
            // Pass multi-element allocation directly to free.
            free(ptr);
        }
        else
        {
            // Place memory on free list.
            const size_t sToWrapper = offsetof(Wrapper, value);
            Wrapper* pWrapper = (Wrapper*)((char*)ptr - sToWrapper);
            pWrapper->pNext = pFree;
            pFree = pWrapper;
        }
    }

    size_type max_size() const throw()
    { return size_t(-1) / sizeof(value_type);  }

    void construct(pointer p, const TT& val)
    { new(static_cast<void*>(p)) TT(val);  }

    void construct(pointer p)
    { new(static_cast<void*>(p)) TT(); }

    void destroy(pointer p)
    { p->~TT(); }
};

///////////////////////////////////////////////////////////////////////////////
// Custom allocator - counts memory allocation calls.
template <class TT>
class CountAlloc : public IAlloc
{
public:
    typedef size_t          size_type;
    typedef ptrdiff_t       difference_type;
    typedef TT*             pointer;
    typedef const TT*       const_pointer;
    typedef TT&             reference;
    typedef const TT&       const_reference;
    typedef TT              value_type;   

    template <class UU>    
    struct rebind
    {
        typedef CountAlloc<UU> other;
    };
 
    struct SharedInfo
    {
    public:
        SharedInfo() : allocCnt(0), refCnt(1)
        {  }
            
        size_t  allocCnt;
        int     refCnt;
    };
    SharedInfo* pSharedInfo;

    CountAlloc () throw() : pSharedInfo(new SharedInfo())
    {  }

    CountAlloc (const CountAlloc& other) : pSharedInfo(other.pSharedInfo)
    { pSharedInfo->refCnt++; }

    template <class U> 
    CountAlloc(const CountAlloc<U>& other) : pSharedInfo((SharedInfo*)other.pSharedInfo)
    { pSharedInfo->refCnt++;}

    bool operator==(const CountAlloc& other) 
    { return this == &other;  } 

    ~CountAlloc () 
    { if (--pSharedInfo->refCnt == 0) delete pSharedInfo;  }

    size_t GetAllocCnt() const
    { return pSharedInfo->allocCnt; }

#if  _MSC_VER==1500
    // VS2008
    pointer address(reference x) const
    { return &x; }

    const_pointer address(const_reference x) const
    { return &x; }
#else
    pointer address(reference _Val) const 
    {	// return address of mutable _Val
        return (_STD addressof(_Val));
    }

    const_pointer address(const_reference _Val) const 
    {	// return address of nonmutable _Val
        return (_STD addressof(_Val));
    }
#endif

    pointer allocate(size_type size, typename CountAlloc<TT>::const_pointer hint = 0)
    {
        pSharedInfo->allocCnt++;
        // return static_cast<pointer>(::malloc(size * sizeof(TT)));   
        return static_cast<pointer>(::operator new(size * sizeof(TT)));
    }

    void deallocate(pointer p, size_type n)
    {
        // ::free(p);
        ::operator delete(p);
    }

    size_type max_size() const throw()
    { return size_t(-1) / sizeof(value_type); }

    void construct(pointer p, const TT& val)
    { new(static_cast<void*>(p)) TT(val); }

    void construct(pointer p)
    { new(static_cast<void*>(p)) TT(); }

    void destroy(pointer p)
    { p->~TT(); }
};

///////////////////////////////////////////////////////////////////////////////
/// Exercise ordered containers 
///
/// 1. Add words to container until it has 'groupSize' elements.
/// 2. Scan remainder of words counting duplicate.
/// 3. When scanning sLookupScale * groupSize words, clear list and goto #1
/// 4. Grab memory allocation count
/// 5. Compute total of all duplicates (used to compare test results)

template <typename Container_t>
size_t Test(const Words& words, size_t groupSize, const char* objName, size_t& mallocCnt)
{
    Container_t container;

    size_t lookupCnt = 0;
    size_t cnt = words.size();

#ifdef USE_COUNTALLOC
    // One shot pre-allocation value use to allocate next set of data nodes.
    // Afterwards it goes back to sAllocRate which is 16.
    container.get_allocator().SetPrealloc(groupSize);
#endif

    typename Container_t::iterator found;
    for (size_t idx = 0; idx < cnt; idx++)
    {
        const std::string* pWord = &(words[idx]);
        if (container.size() < groupSize)
        {
            // Fill - find dup and count, else add new word.
            found = container.find(*pWord);
            if (found != container.end())
                found->second++;    // Count duplicates
            else
            {
                PairSI pairValue(*pWord, 1);
                container.insert(pairValue);
            }
        }
        else
        {
            // Find dup and count, else ignore.
            found = container.find(*pWord);
            if (found != container.end())
                found->second++;    // Only count duplicates

            // Occassionally empty list and repeat fill cycle.
            if (++lookupCnt == groupSize * sLookupScale)
            {
                lookupCnt = 0;
                container.clear();
            }
        }
    }


#ifdef USE_COUNTALLOC
    mallocCnt = container.get_allocator().GetAllocCnt();
#endif

    // The following is diagnostic code to sum word count and return its value
    // to compare to other containers to make sure all instances return the same value.
    size_t total = 0;
    for (typename Container_t::const_iterator iter = container.begin(); iter != container.end(); ++iter)
    {
        total += iter->second;
    }

    return total;
}

///////////////////////////////////////////////////////////////////////////////
/// Exercise ordered containers (with No Count of memory allocations)
///
/// 1. Add words to container until it has 'groupSize' elements.
/// 2. Scan remainder of words counting duplicate.
/// 3. When scanning sLookupScale * groupSize words, clear list and goto #1

template <typename Container_t>
size_t Testnc(const Words& words, size_t groupSize, const char* objName, size_t& mallocCnt)
{
    Container_t container;

    size_t lookupCnt = 0;
    size_t cnt = words.size();

    typename Container_t::iterator found;
    for (size_t idx = 0; idx < cnt; idx++)
    {
        const std::string* pWord = &(words[idx]);
        if (container.size() < groupSize)
        {
            // Fill - find dup and count, else add new word.
            found = container.find(*pWord);
            if (found != container.end())
                found->second++;    // Count duplicates
            else
            {
                PairSI pairValue(*pWord, 1);
                container.insert(pairValue);
            }
        }
        else
        {
            // Find dup and count, else ignore.
            found = container.find(*pWord);
            if (found != container.end())
                found->second++;    // Only count duplicates

            // Occassionally empty list and repeat fill cycle.
            if (++lookupCnt == groupSize * sLookupScale)
            {
                lookupCnt = 0;
                container.clear();
            }
        }
    }

    // Using external Allocator which does not provide Allocation Counts.
    mallocCnt = 0;

    // The following is diagnostic code to sum word count and return its value
    // to compare to other containers to make sure all instances return the same value.
    size_t total = 0;
    for (typename Container_t::const_iterator iter = container.begin(); iter != container.end(); ++iter)
    {
        total += iter->second;
    }

    return total;
}

using namespace std;

#if  _MSC_VER==1500
// VS2008
using namespace stdext;
#endif

///////////////////////////////////////////////////////////////////////////////
template <typename Vec_t>
inline int find(const Vec_t& tbl, const std::string& key)
{
    for (size_t idx=0; idx < tbl.size(); idx++)
    {
        if (tbl[idx].first == key)
            return (int)idx;
    }
    return -1;
}

// Binary search use on non-sorted containers (vector)
///////////////////////////////////////////////////////////////////////////////
template <typename vPair>
int binary_search(const vPair& sortedVecPair, const std::string& key)
{
    // Keep halfing the search space until we reach the end of the vector
    int middle;
    int begin = 0;
    int end = (int)sortedVecPair.size();
    int NotFound = -1;

    while (begin < end)
    {
        // Find the median value between the iterators
        middle = begin + (end - begin) / 2;

        // Re-adjust the iterators based on the median value
        if (sortedVecPair[middle].first == key)
        {
            return middle;
        }
        else if (sortedVecPair[middle].first > key)
        {
            end = middle;
        }
        else
        {
            begin = middle + 1;
        }
    }

    return NotFound;
}

///////////////////////////////////////////////////////////////////////////////
bool vecPairCmp(const PairSI& p1, const PairSI& p2)
{
    return p1.first < p2.first;
}

// Custom tester to count duplicates in a vector (find, upd if found, else add)
///////////////////////////////////////////////////////////////////////////////
template <typename Container_t>
size_t TestVector(const Words& words, size_t groupSize, const char* name, size_t& mallocCnt)
{
    Container_t vec;

    size_t lookupCnt = 0;
    size_t cnt = words.size();
    vec.reserve(groupSize);

    for (size_t idx = 0; idx < cnt; idx++)
    {
        const std::string& word = words[idx];

        if (vec.size() < groupSize)
        {
            int foundIdx = find(vec, word);
            if (foundIdx == -1)
            {
                vec.push_back(PairSI(word, 1));
            }
            else
            {
                vec[foundIdx].second++;
            }
        }
        else
        {
            int foundIdx = find(vec, words[idx]);
            if (foundIdx != -1)
                vec[foundIdx].second++;
            if (++lookupCnt == groupSize * sLookupScale)
            {
                lookupCnt = 0;
                vec.clear();
            }
        }
    }

#ifdef USE_COUNTALLOC
    mallocCnt = vec.get_allocator().GetAllocCnt();
#endif

    // The following is diagnostic code to sum word count and return its value
    // to compare to other containers to make sure all instances return the same value.
    size_t total = 0;
    for (typename Container_t::const_iterator iter = vec.begin(); iter != vec.end(); ++iter)
    {
        total += iter->second;
    }

    return total;
}

// Custom vector tester to sort when full and use binary search to update dups.
///////////////////////////////////////////////////////////////////////////////
template <typename Container_t>
size_t TestSortedVector(const Words& words, size_t groupSize, const char* name, size_t& mallocCnt)
{
    Container_t svec;

    size_t lookupCnt = 0;
    size_t cnt = words.size();
    svec.reserve(groupSize);

    for (size_t idx = 0; idx < cnt; idx++)
    {
        const std::string& word = words[idx];

        if (svec.size() < groupSize)
        {
            int foundIdx = find(svec, word);
            if (foundIdx == -1)
            {
                svec.push_back(PairSI(word, 1));
                if (svec.size() == groupSize)
                    sort(svec.begin(), svec.end(), vecPairCmp);
            }
            else
            {
                svec[foundIdx].second++;
            }
        }
        else
        {
            int foundIdx = binary_search(svec, word);
            if (foundIdx != -1)
                svec[foundIdx].second++;
            if (++lookupCnt == groupSize * sLookupScale)
            {
                lookupCnt = 0;
                svec.clear();
            }
        }
    }

#ifdef USE_COUNTALLOC
    mallocCnt = svec.get_allocator().GetAllocCnt();
#endif

    // The following is diagnostic code to sum word count and return its value
    // to compare to other containers to make sure all instances return the same value.
    size_t total = 0;
    for (typename Container_t::const_iterator iter = svec.begin(); iter != svec.end(); ++iter)
    {
        total += iter->second;
    }

    return total;
}

// Custom tester for Deque to use either front or back interface.
///////////////////////////////////////////////////////////////////////////////
template <typename Container_t>
size_t TestDeque(const Words& words, size_t groupSize, size_t& mallocCnt, bool addToBack)
{
    Container_t queue;

    size_t lookupCnt = 0;
    size_t cnt = words.size();

    for (size_t idx = 0; idx < cnt; idx++)
    {
        const std::string& word = words[idx];

        if (queue.size() < groupSize)
        {
            int foundIdx = find(queue, word);
            if (foundIdx == -1)
            {
#ifdef _DEQUE_EMPLACE
                if (addToBack)
                    queue.emplace_front(word, 1);
                else
                    queue.emplace_back(word, 1);
#else
                if (addToBack)
                    queue.push_back(PairSI(word, 1));
                else
                    queue.push_front(PairSI(word, 1));
#endif
            }
            else
            {
                queue[foundIdx].second++;
            }
        }
        else
        {
            int foundIdx = find(queue, words[idx]);
            if (foundIdx != -1)
                queue[foundIdx].second++;
            if (++lookupCnt == groupSize * sLookupScale)
            {
                lookupCnt = 0;
                queue.clear();
            }
        }
    }

#ifdef USE_COUNTALLOC
    mallocCnt = queue.get_allocator().GetAllocCnt();
#endif

    // The following is diagnostic code to sum word count and return its value
    // to compare to other containers to make sure all instances return the same value.
    size_t total = 0;
    for (typename Container_t::const_iterator iter = queue.begin(); iter != queue.end(); ++iter)
    {
        total += iter->second;
    }

    return total;
}

///////////////////////////////////////////////////////////////////////////////
template <typename Container_t>
size_t TestDequeBack(const Words& words, size_t groupSize, const char* name, size_t& mallocCnt)
{
    const bool isBack = true;
    return TestDeque<Container_t>(words, groupSize, mallocCnt, isBack);
}

///////////////////////////////////////////////////////////////////////////////
template <typename Container_t>
size_t TestDequeFront(const Words& words, size_t groupSize, const char* name, size_t& mallocCnt)
{
    const bool isfront = false;
    return TestDeque<Container_t>(words, groupSize, mallocCnt, isfront);
}

///////////////////////////////////////////////////////////////////////////////
/// Extract words from text file.
void LoadWords(const char* worldFile, Words& words, unsigned maxWordLength)
{
    std::ifstream in(worldFile);
    if (!in.good())
    {
        perror(worldFile);
        std::cerr << "Run program with first argument pointing to a text file which contains 1000+ words.\n"
            "For example, point it to the source code for this program.\n"
            "       perf-hash-map  perf-hash-map.cpp\n";
        exit(0);
    }

    Words wordsFromFile;
    while (in.good())
    {
        std::string word;
        in >> word;
// #define FIXED_SIZE_WORDS
#ifdef FIXED_SIZE_WORDS
        // Force all words to fixed size and semi-unique.
        if (word.empty() == false)
        {
            static char c='a';
            word.resize(maxWordLength-1, c);
            c = ((c + 11) - 'a') % 26 + 'a';
        }
#endif
        if (!word.empty() && word.length() < maxWordLength)
        {
            wordsFromFile.push_back(word);
        }
    }

    
    // Repeat adding words until we have enough.
    do
    {
        words.insert(words.end(), wordsFromFile.begin(), wordsFromFile.end()); 
    } while (words.size() < 40000);
}

///////////////////////////////////////////////////////////////////////////////
bool fileexists(const char* filename)
{
    std::ifstream in(filename);
    return in.good();
}

///////////////////////////////////////////////////////////////////////////////
typedef size_t (*fptr)(const Words&, size_t grp, const char* name, size_t& mallocCnt);

struct ThreadInfo
{
public: 
    ThreadInfo(fptr funcPtr, const Words& words, size_t groupSize) :
        m_funcPtr(funcPtr), m_words(words), m_groupSize(groupSize)
    {  }

    fptr         m_funcPtr;
    const Words& m_words;
    size_t       m_groupSize;
    size_t       m_mallocCnt;
};

static ThreadInfo* sThreadInfoList[sMaxThreads];
static Thread_t sThreadHndList[sMaxThreads];

///////////////////////////////////////////////////////////////////////////////
Thread_return_t WINAPI BackgroundThread(void* who_data)
{
    long threadIdx = (long)who_data;
    ThreadInfo& threadInfo = *sThreadInfoList[threadIdx];
    size_t total = threadInfo.m_funcPtr(threadInfo.m_words, threadInfo.m_groupSize, "xx", threadInfo.m_mallocCnt);
	pthread_exit((Thread_return_t)total);
}

///////////////////////////////////////////////////////////////////////////////
void StartBackgroundThreads(
    unsigned threadCnt, 
    fptr func, 
    const Words& words, 
    size_t groupSize)
{
    if (threadCnt > sMaxThreads)
    {
        std::cerr << "Invalid thread count:" << threadCnt << ", max allowed is " << sMaxThreads << std::endl;
        exit(-1);
    }

    for (unsigned threadIdx=0; threadIdx < threadCnt; threadIdx++)
	{
        sThreadInfoList[threadIdx] = new ThreadInfo(func, words, groupSize);

#ifdef WIN32
		DWORD thread_id;
         if ((sThreadHndList[threadIdx] = CreateThread(NULL, 0, 
             (LPTHREAD_START_ROUTINE)BackgroundThread, (PVOID)threadIdx, 0, &thread_id)) == NULL)
#else
        
        pthread_attr_t* pthread_attr_default = 0;
        if (pthread_create(&(sThreadHndList[threadIdx]), pthread_attr_default,
                    &BackgroundThread, (void*)threadIdx ))
#endif
		{
            std::cerr << "Thread start failed.\n";
            return;
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
void WaitForThreadsToEnd()
{
    // Wait for threads to finish
    for (int threadIdx=0; threadIdx < sThreadCnt; threadIdx++)
    {
#ifdef WIN32
		WaitForSingleObject(sThreadHndList[threadIdx], INFINITE);
#else
        pthread_join(sThreadHndList[threadIdx], NULL);
#endif
    }
}

///////////////////////////////////////////////////////////////////////////////
struct Results
{
public:
    Results() : m_elapsed(0), m_funcValue(0), m_mallocCnt(0)
    { }

    Elapsed_t m_elapsed;
    size_t    m_funcValue;
    size_t    m_mallocCnt;
    string    m_name;
};

///////////////////////////////////////////////////////////////////////////////
Elapsed_t TimeIt(
    Results& results,
    unsigned threadCnt,
    fptr testFunc,
    const Words& words,
    size_t groupSize,
    const char* name)
{
    if (threadCnt > sMaxThreads)
    {
        std::cerr << "Invalid thread count:" << threadCnt << ", max allowed is " << sMaxThreads << std::endl;
        exit(-1);
    }

    // Time foreground function time while running similar function in background.

    ticks startTick = getticks();

    StartBackgroundThreads(threadCnt, testFunc, words, groupSize);

    results.m_funcValue = testFunc(words, groupSize, name, results.m_mallocCnt);
    results.m_name = name;
    Elapsed_t deltaTicks = elapsed(getticks(), startTick);

    WaitForThreadsToEnd();

    results.m_elapsed += deltaTicks;
    return deltaTicks;
}

// Base class to manage running tests and reporting results.
///////////////////////////////////////////////////////////////////////////////
class Tester
{
public:
    static const unsigned sMaxGroups = 10;
    static const unsigned sMaxTests = 20;
    Results results[sMaxGroups][sMaxTests];
    size_t grpSize[sMaxGroups];
    unsigned grpCnt;
    string testHeader;

    void TestAndReport(
        Words& words,
        unsigned threadCnt = 1,
        unsigned repeatCnt = 1,
        bool showMemoryCounts = true);

    virtual 
    unsigned RunTest(Words& words, unsigned threadCnt=1, unsigned repeatCnt=1) = 0;
};

///////////////////////////////////////////////////////////////////////////////
void Tester::TestAndReport(
    Words& words,
    unsigned threadCnt,
    unsigned repeatCnt,
    bool showMemoryCounts)
{
    // Create several group sizes (group size is maximum number unique words to track)
    grpCnt = 0;
    grpSize[grpCnt++] = 100;
    grpSize[grpCnt++] = 500;
    grpSize[grpCnt++] = 1000;

    if (threadCnt > sMaxThreads)
    {
        std::cerr << "Invalid thread count:" << threadCnt << ", max allowed is " << sMaxThreads << std::endl;
        exit(-1);
    }

    unsigned testCnt = RunTest(words, threadCnt, repeatCnt);

    // Build column headings
    string columHeader = " Group Size";
    for (unsigned tstIdx = 0; tstIdx < testCnt; tstIdx++)
    {
        columHeader += ",";
        string name = results[0][tstIdx].m_name;
        columHeader.resize((tstIdx + 2) * 11 - name.length(), ' ');
        columHeader += name;
    }

    // Find minimum time (fastest results) of repeat average.
    Elapsed_t minTm = results[0][0].m_elapsed;
    for (unsigned g = 0; g < grpCnt; g++)
    {
        for (unsigned t = 0; t < testCnt; t++)
        {
            results[g][t].m_elapsed /= repeatCnt;
            minTm = std::min(minTm, results[g][t].m_elapsed);
        }
    }

    // Display results, scaling by minimum time.
    std::cout << testHeader << " Threads:" << threadCnt << ", Repeat : " << repeatCnt << std::endl;
    std::cout << "Results reported as ratio of elapsedTicks/fastestTicks(" << minTm << "):\n";
    std::cout << columHeader << std::endl;
    for (unsigned grpIdx = 0; grpIdx < grpCnt; grpIdx++)
    {
        size_t funcValue = results[grpIdx][0].m_funcValue;
        size_t grpSz = grpSize[grpIdx];
        std::cout
            << std::fixed << std::setprecision(0)
            << std::setw(11) << grpSz
            << std::fixed << std::setprecision(2);

        for (unsigned tstIdx = 0; tstIdx < testCnt; tstIdx++)
        {
            std::cout << "," << std::setw(9) << results[grpIdx][tstIdx].m_elapsed / minTm;
            std::cout << ((results[grpIdx][tstIdx].m_funcValue == funcValue) ? ' ' : '?');
        }
        std::cout << std::endl;
    }
    std::cout << std::endl;

    if (showMemoryCounts)
    {
        std::cout << "Number of Memory Allocations made, malloc() call count:\n";
        std::cout << columHeader << std::endl;
        for (unsigned grpIdx = 0; grpIdx < grpCnt; grpIdx++)
        {
            size_t funcValue = results[grpIdx][0].m_funcValue;
            size_t grpSz = grpSize[grpIdx];
            std::cout
                << std::fixed << std::setprecision(0)
                << std::setw(11) << grpSz
                << std::fixed << std::setprecision(2);

            for (unsigned tstIdx = 0; tstIdx < testCnt; tstIdx++)
            {
                std::cout << "," << std::setw(9) << results[grpIdx][tstIdx].m_mallocCnt;
                std::cout << " ";
            }
            std::cout << std::endl;
        }
        std::cout << std::endl;
    }
}

///////////////////////////////////////////////////////////////////////////////
class TestStl : public Tester
{
    unsigned RunTest(Words& words, unsigned threadCnt = 1, unsigned repeatCnt = 1)
    {
        // Define custom allocators
        typedef CountAlloc<PairSI> CountAllocSI;
        typedef PoolAlloc<PairSI> PoolAllocSI;

#ifdef WIN32
        typedef std::less<std::string> LessStr;
        typedef hash_compare<std::string, LessStr> HashCmp;
        typedef hash_map<std::string, int, HashCmp, CountAllocSI > HashTbl;
        typedef hash_map<std::string, int, hash_compare<std::string, std::less<std::string>>, PoolAllocSI > HashAllocTbl;
        typedef hash_multimap<std::string, int, hash_compare<std::string, std::less<std::string>>, CountAllocSI > HashMmapTbl;
#endif
        typedef std::map<std::string, int, std::less<std::string>, CountAllocSI > MapTbl;
        typedef std::map<std::string, int, std::less<std::string>, PoolAllocSI > MapAllocTbl;
        typedef std::multimap<std::string, int, std::less<std::string>, CountAllocSI > MmapTbl;
        typedef std::vector<PairSI, CountAllocSI> VecTbl;
        typedef std::deque<PairSI, CountAllocSI> DqueTbl;

    if (threadCnt > sMaxThreads)
    {
        std::cerr << "Invalid thread count:" << threadCnt << ", max allowed is " << sMaxThreads << std::endl;
        exit(-1);
    }

        testHeader = "===[ STD/STL ]===\n";
        unsigned testCnt = 0;
        for (unsigned repIdx = 0; repIdx < repeatCnt; repIdx++)
        {
            for (unsigned grpIdx = 0; grpIdx < grpCnt; grpIdx++)
            {
                size_t grpSz = grpSize[grpIdx];
                testCnt = 0;
#ifdef WIN32
                TimeIt(results[grpIdx][testCnt++], threadCnt, Test<HashTbl>, words, grpSz, "HashMap");
                TimeIt(results[grpIdx][testCnt++], threadCnt, Test<HashAllocTbl>, words, grpSz, "HasMpAlloc");
                TimeIt(results[grpIdx][testCnt++], threadCnt, Test<HashMmapTbl>, words, grpSz, "HashMmap");
#endif
                TimeIt(results[grpIdx][testCnt++], threadCnt, Test<MapTbl>, words, grpSz, "Map");
                TimeIt(results[grpIdx][testCnt++], threadCnt, Test<MapAllocTbl>, words, grpSz, "MapAlloc");
                TimeIt(results[grpIdx][testCnt++], threadCnt, Test<MmapTbl>, words, grpSz, "MultiMap");
                TimeIt(results[grpIdx][testCnt++], threadCnt, TestVector<VecTbl>, words, grpSz, "Vector");
#ifdef WIN32
                TimeIt(results[grpIdx][testCnt++], threadCnt, TestSortedVector<VecTbl>, words, grpSz, "SortedVec");
#endif
                TimeIt(results[grpIdx][testCnt++], threadCnt, TestDequeFront<DqueTbl>, words, grpSz, "DequeFront");
                TimeIt(results[grpIdx][testCnt++], threadCnt, TestDequeBack<DqueTbl>, words, grpSz, "DequeBack");
            }
        }

        return testCnt;
    }
};

///////////////////////////////////////////////////////////////////////////////
class TestBoost : public Tester
{
    unsigned RunTest(Words& words, unsigned threadCnt, unsigned repeatCnt)
    {
        // Define allocators 
        // Boost wants some allocators to be const and others not.
        typedef CountAlloc<PairSIconst> CountAllocSIC;
        typedef CountAlloc<PairSI> CountAllocSI;

#ifdef TEST_BOOST_ALLOCATOR
        // Crashes in x86, fails to link in x64
        typedef boost::pool_allocator<PairSI> BoostPoolAllocSI;
        typedef boost::container::map< std::string, int, std::less<const std::string>, BoostPoolAllocSI > MapBAllocTbl;
#endif

        typedef std::vector<PairSI, CountAllocSIC> VecStd;
        typedef boost::container::map< std::string, int, std::less<const std::string>, CountAllocSIC > MapTbl;
        typedef boost::container::multimap< std::string, int, std::less<const std::string>, CountAllocSIC > MMapTbl;
        typedef boost::container::flat_map<std::string, int, std::less<std::string>, CountAllocSI> FlatMapTbl;
        typedef boost::container::vector<PairSI, CountAllocSI> VecTbl;
        typedef boost::container::deque<PairSI, CountAllocSI> DqueTbl;

        testHeader = "===[ Boost " BOOST_LIB_VERSION " ] ===\n";
        unsigned testCnt = 0;
        for (unsigned repIdx = 0; repIdx < repeatCnt; repIdx++)
        {
            for (unsigned grpIdx = 0; grpIdx < grpCnt; grpIdx++)
            {
                size_t grpSz = grpSize[grpIdx];
                testCnt = 0;

#ifdef ALL_STD_VEC_FOR_RESULT_COMPARISON
                TimeIt(results[grpIdx][testCnt++], threadCnt, TestVector<VecStd>, words, grpSz, "stdVecmap");
#endif
                TimeIt(results[grpIdx][testCnt++], threadCnt, Test<MapTbl>, words, grpSz, "map");
#ifdef TEST_BOOST_ALLOCATOR
                TimeIt(results[grpIdx][testCnt++], threadCnt, Testnc<MapBAllocTbl>, words, grpSz, "mapBAlloc");
#endif
                TimeIt(results[grpIdx][testCnt++], threadCnt, Test<FlatMapTbl>, words, grpSz, "flat_map");
                TimeIt(results[grpIdx][testCnt++], threadCnt, Test<MMapTbl>, words, grpSz, "MuliMap");
                TimeIt(results[grpIdx][testCnt++], threadCnt, TestVector<VecTbl>, words, grpSz, "Vector");
                TimeIt(results[grpIdx][testCnt++], threadCnt, TestDequeFront<DqueTbl>, words, grpSz, "DequeFront");
                TimeIt(results[grpIdx][testCnt++], threadCnt, TestDequeBack<DqueTbl>, words, grpSz, "DequeBack");
            }
        }

        return testCnt;
    }
};

// Display summary of word properties.
///////////////////////////////////////////////////////////////////////////////
void CheckWords(const Words& words)
{
    typedef std::map<std::string, int> Freq_t;
    Freq_t freq;
    Freq_t::const_iterator iter;
    size_t totalLength = 0;
    size_t maxLength = 0;

    size_t cnt = words.size();
    for (size_t idx = 0; idx < cnt; idx++)
    {
        const std::string& word = words[idx];
        iter = freq.find(word);
        if (iter == freq.end())
            freq[words[idx]] = 1;
        else
            freq[words[idx]]++;

        totalLength += word.length();
        maxLength = std::max(maxLength, word.length());
    }

    std::cout << "\nInput Word Analysis:\n";
    std::cout << std::fixed << std::setprecision(2);
    std::cout << std::setw(10) << words.size() << " words\n";
    std::cout << std::setw(10) << freq.size() << " unique words\n";
    std::cout << std::setw(10) << words.size() / freq.size()  << " average #duplicates\n";
    std::cout << std::setw(10) << totalLength / words.size() << " average length\n";
    std::cout << std::setw(10) << maxLength << " maximum length\n";
    std::cout << endl;
    std::cout << std::setw(10) << CompilerVersion << " Compiler\n";
    std::cout << std::setw(10) << BuildTarget << " BuildTarget\n";      
    std::cout << std::setw(10) << BuildDesc << " BuildType\n";
    std::cout << std::setw(10) << CharSet << " CharacterSet\n";
}

///////////////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
    // Load a text file and extract words from it.
    // 
    // 1. Programs first argument can be a file to load
    //      perf-hash-map  <words.txt>
    // 2. If no argument, try and load input-words.txt
    //     If executable path fails, go up one directory.
    //        ..\input-words.txt
    //
    char filenameBuf[256];
    char* wordFilename = filenameBuf;
    if (argc > 1)
        wordFilename = argv[1];
    else
    {
        // wordFilename = __FILE__ ;
        wordFilename = "input-words.txt";
        if (!fileexists(wordFilename))
        {
            std::cerr << "  Failed to open file:" << wordFilename << "\n";

            strncpy(filenameBuf, argv[0], sizeof(filenameBuf));
            strcpy(strrchr(filenameBuf, '.'), ".cpp");
            if (!fileexists(filenameBuf))
            {
                std::cerr << "  Failed to open file:" << wordFilename << "\n";

                char dirSlash = ((strchr(filenameBuf, '\\') != NULL) ? '\\' : '/');
                char * dir1 = strrchr(filenameBuf, dirSlash);
                if (dir1 && dir1 > filenameBuf)
                {
                    *dir1-- = '\0';
                    char * dir2 = strrchr(filenameBuf, dirSlash);
                    dir1[1] = dirSlash;
                    strcpy(dir2, dir1);
                }
            }
        }
    }

    // Output banner.
    std::cout << "\nC++ Standard Template Library and Boost Performance\n"
        "Program created by: Dennis Lang http://home.comcast.net/~lang.dennis/index.html\n\n";

    Words words;
    const int maxWordLength = 40;
    LoadWords(wordFilename, words, maxWordLength);
    CheckWords(words);

#if 0
    // No background thread, single pass, good for debugging.
    TestHashMapVector(words, 0, 1, showMemoryCounts);
#else
    // Four background threads, 4 passes.
    TestStl testStl;
    testStl.TestAndReport(words, sThreadCnt, sRepeatCnt, showMemoryCounts);

    TestBoost testBoost;
    testBoost.TestAndReport(words, sThreadCnt, sRepeatCnt, showMemoryCounts);
#endif

    return 0;
}
